// This code is designed to take a "crunched" PSF and prepare files for use in an 
// MLEM/OSEM reconstruction algorithm.  The naming structure for the files is married to the
// reconstruction code so be wary of altering file names.

// jyh,11/05

// Files:
// coords --> the coordinates for the PSF
// sn --> the sensitivity map
// num_vals --> the number of non-zero values at each coordinate
// num_vals_sum --> a running sum of the non-zero values
// pos --> the position of each non-zero values
// counts --> the number of counts at each position

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

main(int argc, char **argv)
{

  int cam_num, mdrf;
  long int num_coords;
  int i,j,n;
  short int *num_vals,*coords,*pos_chunk, *counts_chunk;
  long int *num_vals_sum;
  float *sn;
  FILE *psf_file, *coords_file, *pos_file, *counts_file, *num_vals_file, *num_vals_sum_file, *sn_file;
  char *path, *psf_in, filename[256];
  short int data[4];
  int sn_temp;
  long int sum_temp;
  long int max_vector,counter,step;

  setlinebuf(stdout);
 
  // Check the number of input arguments
  // The input should include a path to the input psf file (after crunching) and for the output files, the name of the input 
  // psf file, the camera number, the number of coordinates, and the mdrf dimension.  
  if (argc != 6){
    printf("usage:prog_name path psf_in cam_num num_coords mdrf\n");
    printf("e.g. ./psf_sort ../Data/PSF_10_05/ psf_10_05_cam_01 1 41888 79\n");
    exit(0);
  }

  path = argv[1];
  psf_in = argv[2];
  cam_num = atoi(argv[3]);
  num_coords = atoi(argv[4]);
  mdrf = atoi(argv[5]);

  // OPEN PSF FILE
  sprintf(filename,"%s%s",path,psf_in);
  psf_file = fopen(filename, "r");
  if (psf_file <= 0){
    printf("couldn't open %s\n", psf_in);
    exit(0);
  }
  else printf(" opened the psf file: %s\n",psf_in);

  // SEVERAL FILES WILL NOW BE OPENED.  THE PSF WILL BE SPLIT INTO THESE FILES.  THIS SORTING
  // IS DONE TO IMPROVE RECONSTRUCTION TIMES.
  
  // COORDINATES FILE - WILL CONTAIN THE PSF COORDINATES
  sprintf(filename,"%scoords_0%d",path,cam_num);
  coords_file = fopen(filename,"w");
  if (coords_file <= 0){
    printf("couldn't open %s\n", coords_file);
    exit(0);
  }

  // NUM_VALS FILE - WILL CONTAIN THE NUMBER OF NON-ZERO VALUES AT EACH COORDINATE
  sprintf(filename,"%snum_vals_0%d",path,cam_num);
  num_vals_file = fopen(filename,"w");
  if (num_vals_file <= 0){
    printf("couldn't open %s\n", num_vals_file);
    exit(0);
  }

  // NUM_VALS_SUM FILE - WILL CONTAIN A RUNNING SUM OF ALL NON-ZERO VALUES
  sprintf(filename,"%snum_vals_sum_0%d",path,cam_num);
  num_vals_sum_file = fopen(filename,"w");
  if (num_vals_sum_file <= 0){
    printf("couldn't open %s\n", num_vals_sum_file);
    exit(0);
  }

  // POS FILE - WILL CONTAIN THE CAMERA FACE LOCATIONS FOR THE NON-ZERO VALUES
  sprintf(filename,"%spos_0%d",path,cam_num);
  pos_file = fopen(filename,"w");
  if (pos_file <= 0){
    printf("couldn't open %s\n", pos_file);
    exit(0);
  }

  // COUNTS FILE - WILL CONTAIN THE ACTUAL VALUES FOR EACH NON-ZERO VALUE
  sprintf(filename,"%scounts_0%d",path,cam_num);
  counts_file = fopen(filename,"w");
  if (counts_file <= 0){
    printf("couldn't open %s\n", counts_file);
    exit(0);
  }

  // SENSITIVITY FILE - WILL CONTAIN THE SUM OF THE VALUES OVER THE CAMERA FACE FOR EACH COORDINATE POINT
  sprintf(filename,"%ssn_cam_%d",path,cam_num);
  sn_file = fopen(filename,"w");
  if (sn_file <= 0){
    printf("couldn't open %s\n", sn_file);
    exit(0);
  }

  // WE CAN'T DO ALL OF THE SORTING IN ONE BIG CHUNK NOR DO WE WANT TO FWRITE 
  // 6 FILES 40000 TIMES.  THE FOLLOWING LINES ENABLE US TO ACHIEVE A HAPPY MEDIUM.
  max_vector = 2000000;  // arbitrary but seems to work well
  step = max_vector/(mdrf*mdrf);
  counter = 0;

  // ALLOCATE MEMORY FOR THE FILES
  pos_chunk = (short int*)malloc(max_vector*sizeof(short int));
  counts_chunk = (short int*)malloc(max_vector*sizeof(short int));
  coords = (short int*)malloc(3*num_coords*sizeof(short int));
  num_vals = (short int*)malloc(num_coords*sizeof(short int));
  num_vals_sum = (long int*)malloc(num_coords*sizeof(long int));
  sn = (float *)malloc(num_coords*sizeof(float));

  // BASICALLY, THE CRUNCHED PSF IS WRITTEN OUT IN A SPECIFIED FORMAT.  ALL THIS CODE IS DOING IS READING IN THE
  // CRUNCHED PSF AND SPLITTING IT INTO SEVERAL COMPONENTS
  for (i = 0; i < num_coords; i++){
    if ((i+1)%step == 0){
      for (n = 0; n < counter; n++) counts_chunk[n] = htons(counts_chunk[n]);
      fwrite(counts_chunk,sizeof(short int),counter,counts_file);
      for (n = 0; n < counter; n++) pos_chunk[n] = htons(pos_chunk[n]);
      fwrite(pos_chunk,sizeof(short int),counter,pos_file);
      counter = 0;
      free(pos_chunk);free(counts_chunk);
      pos_chunk = (short int*)malloc(max_vector*sizeof(short int));
      counts_chunk = (short int*)malloc(max_vector*sizeof(short int));
    }

    fread(data,sizeof(short int),4,psf_file);
    for (n = 0; n < 4; n++) data[n] = ntohs(data[n]);
    fread(&pos_chunk[counter],sizeof(short int),data[3],psf_file);
    for (n = 0; n < data[3]; n++) pos_chunk[counter + n] = ntohs(pos_chunk[counter+n]);
    fread(&counts_chunk[counter],sizeof(short int),data[3],psf_file);
    for (n = 0; n < data[3]; n++) counts_chunk[counter + n] = ntohs(counts_chunk[counter+n]);

    sn_temp = 0;
    for (j = counter; j < counter + data[3];j++){
      sn_temp = sn_temp + counts_chunk[j];
    }
    sn[i] = sn_temp;

    counter = counter + data[3];

    for (j = 0; j < 3; j++) coords[j*num_coords+i] = data[j];
    
    num_vals[i] = data[3];    

    if (i > 0){
      sum_temp = 0;
      for (j = 0; j < i; j++) sum_temp = sum_temp + num_vals[j];
      num_vals_sum[i] = sum_temp;
    }
    else num_vals_sum[i] = 0;

    if (i%1000 == 0) printf("%d out of %ld done\n",i,num_coords);

  }

  for (n = 0; n < counter; n++) pos_chunk[n] = htons(pos_chunk[n]);
  fwrite(pos_chunk,sizeof(short int),counter,pos_file);

  for (n = 0; n < counter; n++) counts_chunk[n] = htons(counts_chunk[n]);
  fwrite(counts_chunk,sizeof(short int),counter,counts_file);

  for (n = 0; n < num_coords; n++) *(long *)&sn[n] = htonl(*(long *)&sn[n]);
  fwrite(sn,sizeof(float),num_coords,sn_file);

  for (n = 0; n < num_coords; n++) num_vals[n] = ntohs(num_vals[n]);
  fwrite(num_vals,sizeof(short int),num_coords,num_vals_file);

  for (n = 0; n < num_coords; n++) num_vals_sum[n] = htonl(num_vals_sum[n]);
  fwrite(num_vals_sum,sizeof(long int),num_coords,num_vals_sum_file);

  for (n = 0; n < num_coords*3;n++) coords[n] = htons(coords[n]);
  fwrite(coords,sizeof(short int),num_coords*3,coords_file);

  free(coords);free(num_vals_sum);free(num_vals);free(sn);free(pos_chunk);free(counts_chunk);
}
