/****************************************************
  This is a 2D adaptive window sized ORD-order
  polynomial fit and interpolation algorithm used for 
  multi-element data.  The smallest window size (WMIN)
  is used at the MDRF point with the largest signal (Smax).
  Window size for all other points will scale as:
    min[WMAX, WMIN*(S(x,y,c)/Smax)^0.25]

  The input MDRF is a (NR x NC x NCHN] maxtrix read
  from a input file of data type (float).  If there
  is header information in the input file (FN_IN), it 
  can be skipped (off = # of bytes in header).

  Writing this header information to the output file
  is left to the user to do after the output file has
  been generated by this subroutine.

  An output MDRF of dimensions (NRO x NCO x NCHN) is
  written to the specified output file (FN_OUT).

  Useage:
  ./exec FN_IN OFF NR NC NCHN ORD WMIN WMAX NRO NCO FN_OUT

   William Hunter, Dec,14,2004
*****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#define INF ((long )1e8)

/*PRIMATIVES***************************************/
double determ(double *matin, long k);
void ludcmp(double **a, int n, int *indx, double *d);
double *vector(long nl, long nh);
int *ivector(long nl, long nh);
double **matrix(long nrl, long nrh, long ncl, long nch);
void free_vector(double *v, long nl);
void free_ivector(int *v, long nl);
void free_matrix(double **m, long nrl, long ncl);
void cpy1D2D(double **matout, double *matin, long k);
/***************************************************/


int main( int argc, char **argv )
{
  char   fni[1000];    // Input filename for 2D func. "m" (float)
  FILE   *fpi;         // Input file pointer
  long   off;          // Offset for input file
  long   nr,nc;        // # rows & # columns of input 2D matrix "m"
  long   nro,nco;      // # rows & # columns of output 2D matrix "mf"
  double frac;         // # out/in ratio
  long   chn,nchn;     // # channels (measurements) 
  long   ord;          // Polynomial order for fit.  For example:
                       //   z=f(1,X,...,X^n,Y,YX,...YX^(n-1),Y^2,...,Y^n)
  long   w,wmin,wmax;  // Limits for fit window size (1D size ... window is w*w)
  long   n, nw;        // Fit window size variable and counter
  double *wmap;        // Map of window sizes used
  char   fno[1000];    // Output filename for 2D fitted function "mf" (float)
  FILE   *fpo;         // Output file pointer
  float  *m;           // 2D input function
  double mmin, mmax;   // min and max locally filtered signals
  double msum;         // local sum signal
  double sigmin;       // signal for which maximum window size is used
  float  *mf;          // 2D output fitted function
  long   lwmax,lwmin;  // # of fit points before spline point for min & max window sizes
  long   hwmax,hwmin;  // # of fit points before spline point for min & max window sizes
  long   *lw, *hw;     // # of fit points before & after spline point (in each dimension)
  long   lwx,hwx;      // X-limits for fit window
  long   lwy,hwy;      // Y-limits for fit window
  long   X, Y;         // point subscript within window
  long   k, nk;        // # of terms in polynomial (determined below)
  long   i,j;          // Term indecies
  long   xi,yi;        // Fit point index
  long   xii,yii;      // Spline point index
  long   xo,yo;        // Output index
  long   *xoi,*yoi;    // Output to input cross-reference indecies
  double *dx,*dy;      // Fit to Spline points distance
  long   *XX;          // Polynomial terms over range of fit
  double *Xx,*Xy;      // Monomial terms over range of fit
  float  *p;           // Poly coeff used for spline
  double *D, *Di;      // Meth.of Det., indep-var-func*indep-var-func matrix
  double *DY;          // Meth.of Det., indep-var-func*dep-var vector
  double *A;           // Meth.of Det., coef. solution matrix
  double *dD, dDi;     // Det. of D
  double dA;           // Det. of A
  int n_fix;
  time_t t1;



  time(&t1);
  printf("Start Time: %s\n",asctime(localtime(&t1)));

  
  if (argc != 12) {
    fprintf(stderr,"Usage: %s <path/file_in> <header offset> <# rows in> <# columns in> <#chn> <order> <win min size> <win max size> <# rows out> <# columns out> <path/file_out>\n",argv[0]);
    exit(1);
  }
  
  strcpy(fni,argv[1]);
  off  = ( long )atoi(argv[2]);
  nr   = ( long )atoi(argv[3]);
  nc   = ( long )atoi(argv[4]);
  nchn = ( long )atoi(argv[5]);
  ord  = ( long )atoi(argv[6]);
  wmin = ( long )atoi(argv[7]);
  if (wmin<(ord+1)) {
    fprintf(stderr,"Minimum window fit size must be greater than (ORDER+1)\n");
    exit(1);
  }
  wmax = ( long )atoi(argv[8]);
  if (wmax <= 0) wmax = ((long )INF);
  if (wmax > nr) wmax = nr;
  if (wmax > nc) wmax = nc;
  nro  = ( long )atoi(argv[9]);
  nco  = ( long )atoi(argv[10]);
  strcpy(fno,argv[11]);
  
  /*  if ( (fpo=fopen(fno, "r")) != NULL){
    fclose(fpo);
    fprintf(stderr,"<path/file_out> already exists\n");
    exit(1);
    }*/
  
  if ( (fpi=fopen(fni, "r")) == NULL){
    fprintf(stderr,"<path/file_in> does not exist\n");
    exit(1);
  }
  
  m  = malloc(nr*nc*nchn*sizeof(float));
  fseek(fpi,off,SEEK_SET);
  fread(m, sizeof(float), nc*nr*nchn, fpi);
  for (n_fix = 0; n_fix < nc*nr*nchn; n_fix++) *(long *)&m[n_fix] = ntohl(*(long *)&m[n_fix]);
  fclose(fpi);
  
  
  /* Extent of sub-coordinates for SMALLEST fit window */
  lwmin = (wmin-1)/2;  hwmin = wmin-lwmin-1;
  
  
  /* Loop over input map to find largest and smallest signal */
  wmap = malloc(nr*nc*nchn*sizeof(double));
  mmin = INF; mmax = -INF;
  
  for (chn=0; chn<nchn; chn++) {
    for (xi=lwmin; xi<(nc-hwmin); xi++) for (yi=lwmin; yi<(nr-hwmin); yi++) {
      msum = 0;
      for (X=-lwmin; X<=hwmin; X++) for (Y=-lwmin; Y<=hwmin; Y++)
	msum += m[(yi+Y)+((xi+X)+chn*nc)*nr];
      wmap[yi+(xi+chn*nc)*nr] = msum;
      if (mmax<msum) mmax = msum;
      if (mmin>msum) mmin = msum;
    }
  }
  
  /* Determine maximum window size used */
  sigmin = mmax*pow((((double )wmin)/((double )wmax)),4.0);
  if (mmin>sigmin) {
    msum = ((double)wmin)*sqrt(sqrt(((double)mmax)/((double)mmin)));
    w=msum;  if ((msum-((double )w))>=0.5) w+=1;
    if (w<wmax) wmax = (long)w;
  }
  
  
  /* Precompute polynomial terms for each window size used */
  nk = (long)(ord+1)*(ord+2)/2; /* Number of polynomial terms to fit */
  nw=wmax-wmin+1;               /* Number of window sizes used */
  
  lw  = malloc(nw*sizeof(long));
  hw  = malloc(nw*sizeof(long));
  XX  = malloc(wmax*wmax*nk*sizeof(long));
  Xx  = malloc(nco*nk*sizeof(double));
  Xy  = malloc(nro*nk*sizeof(double));  
  D   = malloc(nk*nk*nw*sizeof(double));
  Di  = malloc(nk*nk*sizeof(double));
  dD  = malloc(nw*sizeof(double));
  A   = malloc(nk*nk*sizeof(double));
  DY  = malloc(nk*sizeof(double));
  p   = malloc(nk*nr*nc*sizeof(double));
  mf  = malloc(nro*nco*nchn*sizeof(float));
  xoi = malloc(nco*sizeof(long));
  yoi = malloc(nro*sizeof(long));
  dx  = malloc(nco*sizeof(double));
  dy  = malloc(nro*sizeof(double));
  
  /*Determine output to input cross-reference index & distances */
  frac = ((double )(nc-1))/((double )(nco-1));
  for (xo=0; xo<nco; xo++) {
    dx[xo]  = ((double )xo)*frac;
    xoi[xo] = (long )dx[xo];
    if ((dx[xo]-(double )xoi[xo])>=0.5) xoi[xo] += 1;
    dx[xo] -= (double )xoi[xo];
    xii = xoi[xo];
    if (xoi[xo]<lwmin) {xoi[xo]=lwmin;}
    else if (xoi[xo]>=(nc-hwmin)) {xoi[xo]=nc-hwmin-1;}
    k=0; 
    for (j=0; j<=ord; j++) for (i=0; i<=(ord-j); i++){
      //Xx[k+xo*nk]=pow(((xii-xoi[xo])+dx[xo]),i);  k+=1;
      Xx[k+xo*nk]=pow(((double )(xii-xoi[xo])+dx[xo]),(double )i);  k+=1;
    }
  }
  
  frac = ((double )(nc-1))/((double )(nco-1));
  for (yo=0; yo<nro; yo++){
    dy[yo]   = ((double )yo)*((double )(nr-1))/((double )(nro-1));
    yoi[yo]  = (long )dy[yo];
    if ((dy[yo]-(double )yoi[yo])>=0.5) yoi[yo] += 1;
    dy[yo]  -= (double )yoi[yo];
    yii = yoi[yo];
    if (yoi[yo]<lwmin) {yoi[yo]=lwmin;} 
    else if (yoi[yo]>=(nr-hwmin)) {yoi[yo]=nr-hwmin-1;}
    k=0; 
    for (j=0; j<=ord; j++) for (i=0; i<=(ord-j); i++){
      //Xy[k+yo*nk]=pow(((yii-yoi[yo])+dy[yo]),j);  k+=1;
      Xy[k+yo*nk]=pow(((double )(yii-yoi[yo])+dy[yo]),(double )j);  k+=1;
    }
  }

    
  /* Extent of sub-coordinates for LARGEST fit window */
  lwmax = (wmax-1)/2;  hwmax = wmax-lwmax-1;
  
  
  k=0;  /* Polynomial terms over fit window coordinates */
  for (j=0; j<=ord; j++)   for (i=0; i<=(ord-j); i++){
    for (X=0; X<wmax; X++) for (Y=0; Y<wmax; Y++)
      XX[Y+(X+k*wmax)*wmax]= (pow((X-lwmax),i) * pow((Y-lwmax),j));
    k += 1;
  }
  
  
  for (n=0; n<nw; n++){
    lw[n] = (long )((double )(wmin+n-1)/2);   /* # points (1D) before spline point */
    hw[n] = wmin+n-lw[n]-1;                   /* # points (1D) after spline point */
    
    /* Independant-variable function-product matrix */
    for (j=0; j<nk; j++)   for (i=0; i<nk; i++) {
      D[i+(j+n*nk)*nk] = 0;
      for (X=0; X<(wmin+n); X++) for (Y=0; Y<(wmin+n); Y++) {
	xii = X+lwmax-lw[n];  yii = Y+lwmax-lw[n];
	D[i+(j+n*nk)*nk] +=  (double )XX[yii+(xii+i*wmax)*wmax]
	                    *(double )XX[yii+(xii+j*wmax)*wmax];
      }
      Di[i+j*nk] = D[i+(j+n*nk)*nk];
    }
    dD[n] = determ(Di,nk);
  }
  
  
  /* For each point further than wmin/2 from boundary,
    compute window size & ORD-order spline */
  for (chn=0; chn<nchn; chn++) {
    
    for (xi=lwmin; xi<(nc-hwmin); xi++) for (yi=lwmin; yi<(nr-hwmin); yi++) {
      msum = wmap[yi+(xi+chn*nc)*nr];            //start with local signal amplitude M
      if (((double)msum)>sigmin) {
	msum = ((double)wmin)*sqrt(sqrt(((double)mmax)/((double)msum))); //find window size relative to Mmax
      } else {
	msum = wmax;
      }
      w=msum;  if ((msum-((double )w))>0.5) w+=1;
      if (w>wmax) w=wmax;
      wmap[yi+(xi+chn*nc)*nr] = w;               //Record array of window sizes used
      n = w-wmin;                                  //Index of window size used for (xi,yi)
      lwx = lw[n]-xi;               if (lwx<0) lwx=0;
      hwx = w - (xi+hw[n]-(nc-1));  if (hwx>w) hwx=w;
      lwy = lw[n]-yi;               if (lwy<0) lwy=0;
      hwy = w - (yi+hw[n]-(nr-1));  if (hwy>w) hwy=w;
      
      if (lwx>0 || lwy>0 || hwx<w || hwy<w) {
	/* Spline window is truncated recompute Independant-variable 
	   function-product matrix for truncated region */
	for (j=0; j<nk; j++) for (i=0; i<nk; i++) {
	  Di[i+j*nk] = 0;
	  for (X=lwx; X<hwx; X++) for (Y=lwy; Y<hwy; Y++) {
	    xii = X+lwmax-lw[n];  yii = Y+lwmax-lw[n];
	    Di[i+j*nk] +=  (double )XX[yii+(xii+i*wmax)*wmax]
	      *(double )XX[yii+(xii+j*wmax)*wmax];
	  }
	  A[i+j*nk] = Di[i+j*nk]; // Meth.of Det., coef. solution matrix
	}
	dDi = determ(Di,nk);
      } else {
	/* Use precomputed values */
	for (j=0; j<nk; j++) for (i=0; i<nk; i++) {
	  Di[i+j*nk] = D[i+(j+n*nk)*nk];
	  A[i+j*nk] = Di[i+j*nk];
	}
	dDi = dD[n];
      }
      
      /* Pre-compute Meth.of Det., indep-var-func*dep-var vector */
      for (j=0; j<nk; j++){
	DY[j]=0;
	for (X=lwx; X<hwx; X++) for (Y=lwy; Y<hwy; Y++) {
	  xii = X+lwmax-lw[n];  yii = Y+lwmax-lw[n];
	  DY[j] += XX[yii+(xii+j*wmax)*wmax]*m[(yi+Y-lw[n])+((xi+X-lw[n])+chn*nc)*nr] ;
	}
      }
      
      /* NOTE: If we want to compute all coefficients needed to fit points
	 near boundary, then use the followin condition. */
      //if (xi==lw[0] || xi==(nc-hw[0]-1) || yi==lw[0] || yi==(nr-hw[0]-1)) {
      
      
      /* Compute ORD-term polynomial coefficients for local spline */
      for (j=0; j<nk; j++) {
	for (i=0; i<nk; i++) A[i+j*nk] = DY[i];      // Replace column of A with DY
	dA = determ(A,nk);
	p[yi+(xi+j*nc)*nr]=(float )(dA/dDi);        // Jth-term coefficient
	for (i=0; i<nk; i++) A[i+j*nk] = Di[i+j*nk]; //Restore A to Di
      }
      
      
      /* We only need 0th-order term for spline of center point */
      //WITHOUT INTERPOLATED POINTS:
      //mf[yi+(xi+chn*nc)*nr]=(double )p[yi+(xi+0*nc)*nr];
      /* NOTE: We can use other order terms to fit points near edge,
	 or to interpolate points between origional point array */
      
    }
    
    
    //WITHOUT INTERPOLATED POINTS:
    /*The fit for points with distance<wmin/2 from edge of M is performed
      using closest WxW fit window further than wmin/2 from edge */
    //for (xi=0; xi<nc; xi++) for (yi=0; yi<nr; yi++){
    //if(xi<lwmin || xi>=(nc-hwmin) || yi<lwmin || yi>=(nr-hwmin)) {
    //xii = xi;   yii = yi;
    //if (xii<lwmin) {xii = lwmin;} else if (xii>=(nc-hwmin)) {xii = nc-hwmin-1;}
    //if (yii<lwmin) {yii = lwmin;} else if (yii>=(nr-hwmin)) {yii = nr-hwmin-1;}
    //mf[yi+(xi+chn*nc)*nr] = 0;
    //for (j=0; j<nk; j++)
    //mf[yi+(xi+chn*nc)*nr] += (double )XX[(lwmax+yi-yii)+((lwmax+xi-xii)+j*wmax)*wmax]
    //                        *(double )p[yii+(xii+j*nc)*nr];
    //} }
    
    
    //Fitting and Interpolation (ORD-order polynomial spline) of output matrix
    for (xo=0; xo<nco; xo++) {
      for (yo=0; yo<nro; yo++){
	mf[yo+(xo+chn*nco)*nro]=0;
	for (k=0; k<nk; k++) {
	  mf[yo+(xo+chn*nco)*nro] += Xx[k+xo*nk]*Xy[k+yo*nk]*(double)p[yoi[yo]+(xoi[xo]+k*nc)*nr];
	}
      }
    }
  }
  
  fpo = fopen(fno, "w");
  for (n_fix = 0; n_fix < nco*nro*nchn; n_fix++) *(long *)&mf[n_fix] = htonl(*(long *)&mf[n_fix]);
  fwrite(mf, sizeof(float), nco*nro*nchn, fpo);
  fclose(fpo);  

  //NOTE: Might also write out Coefficients, P, for use with interpolation algorithm
  //NOTE: Might also write out Window sizes, Wmap, to examine wellness of fit
  
  free(dy);
  free(dx);
  free(yoi);
  free(xoi);
  free(mf);
  free(p);
  free(DY);
  free(A);
  free(dD);
  free(Di);
  free(D);
  free(XX);
  free(Xx);
  free(Xy);
  free(hw);
  free(lw);
  free(wmap);
  free(m);

  time(&t1);
  printf("End Time: %s\n",asctime(localtime(&t1)));
  
  return 0;
}


double determ(double *matin, long k)
/* Compute determinant of matin (a sqare matrix
   listed as a column vector with k*k elements) */
{
  double **mat,d;
  int    *indx;
  long   j;
  
  mat = matrix(1,k,1,k);
  indx = ivector(1,k);
  
  cpy1D2D(mat,matin,k);
  ludcmp(mat,k,indx,&d); // This returns d as +/- 1.
  for(j=1;j<=k;j++) d*=mat[j][j];

  free_ivector(indx,1);
  free_matrix(mat,1,1);
  
  return d;
  
}



void cpy1D2D(double **matout, double *matin, long k)
{
  long i, j;
  
  for (j=0; j<k; j++)
    for (i=0; i<k; i++)
      matout[j+1][i+1]=matin[i+j*k];
}



#define NRANSI
#define TINY 1.0e-20;
/* Code for Lower/Upper triangle decomposition of a
   square matrix.   Obtained from Numerical Recipies */
void ludcmp(double **a, int n, int *indx, double *d) {
  int i,j,k, imax=0;
  double big,dum,sum,temp;
  double *vv;
  
  vv=vector(1,n);
  *d=1.0;
  for (i=1;i<=n;i++) {
    big=0.0;
    for (j=1;j<=n;j++)
      if ((temp=fabs(a[i][j])) > big) big=temp;
    if (big == 0.0) {
      fprintf(stderr,"Singular matrix in routine ludcmp\n");
      exit(EXIT_FAILURE);
    }
    vv[i]=1.0/big;
  }
  for (j=1;j<=n;j++) {
    for (i=1;i<j;i++) {
      sum=a[i][j];
      for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
    }
    big=0.0;
    for (i=j;i<=n;i++) {
      sum=a[i][j];
      for (k=1;k<j;k++)
	sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
      if ( (dum=vv[i]*fabs(sum)) >= big) {
	big=dum;
	imax=i;
      }
    }
    if (j != imax) {
      for (k=1;k<=n;k++) {
	dum=a[imax][k];
	a[imax][k]=a[j][k];
	a[j][k]=dum;
      }
      *d = -(*d);
      vv[imax]=vv[j];
    }
    indx[j]=imax;
    if (a[j][j] == 0.0) a[j][j]=TINY;
    if (j != n) {
      dum=1.0/(a[j][j]);
      for (i=j+1;i<=n;i++) a[i][j] *= dum;
    }
  }
  free_vector(vv,1);
}
#undef TINY
#undef NRANSI



#define NR_END 1
double *vector(long nl, long nh)
/* Allocate a double vector with subscript range v[nl..nh] */
{
  double *v;
  
  v=malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
  if (!v) {
    fprintf(stderr,"allocation failure in vector()\n");
    exit(EXIT_FAILURE);
  }
  return v-nl+NR_END;
}


int *ivector(long nl, long nh)
     /* allocate an int vector with subscript range v[nl..nh] */
{
  int *v;
  v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
  if (!v) {
    fprintf(stderr,"allocation failure in ivector()");
    exit(EXIT_FAILURE);
  }
  return v-nl+NR_END;
}


double **matrix(long nrl, long nrh, long ncl, long nch)
/* Allocate a double matrix with subscript range m[nrl..nrh][ncl..nch] */
{
  long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
  double **m;
  
  /* Allocate pointers to rows */
  m=malloc((size_t)((nrow+NR_END)*sizeof(double*)));
  if (!m) {
    fprintf(stderr,"allocation failure 1 in matrix()\n");
    exit(EXIT_FAILURE);
  }
  m += NR_END;
  m -= nrl;
  
  /* Allocate rows and set pointers to them */
  m[nrl]=malloc((size_t)((nrow*ncol+NR_END)*sizeof(double)));
  if (!m[nrl]) {
    fprintf(stderr,"allocation failure 2 in matrix()\n");
    exit(EXIT_FAILURE);
  }
  m[nrl] += NR_END;
  m[nrl] -= ncl;
  
  for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;
  
  /* Return pointer to array of pointers to rows */
  return m;
}


#define FREE_ARG char*
void free_vector(double *v, long nl)
/* Free a double vector allocated with vector() */
{
  free((FREE_ARG) (v+nl-NR_END));
}


void free_ivector(int *v, long nl)
     /* free an int vector allocated with ivector() */
{
  free((FREE_ARG) (v+nl-NR_END));
}


void free_matrix(double **m, long nrl, long ncl)
/* Free a double matrix allocated by matrix() */
{
  free((FREE_ARG) (m[nrl]+ncl-NR_END));
  free((FREE_ARG) (m+nrl-NR_END));
}
#undef FREE_ARG
#undef NR_END
