/*This program finds the least-square fit of the gaussian
  coefficients of the input image.

  The gaussian coefficients are defined as follows:
  sum:    sum (sensitivity) of the input projection image
  mx, my: centroid in x, y directions
  phi:    rotation angle for principal axes [-PI/2, PI/2]
  lx, ly: stand for lambda_x and lambda_y,
          eigenvalues of the covariance matrix (lx > ly)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "lm.h"

/* 2D Gaussian function to be minimized with Lev-Mar and its jacobian */

void gauss2d(double *p, double *x, int m, int n, void *data)
{
  register int i;
  int j, k;
  double uj, uk;
  int DIM;
  
  DIM = (int) sqrt(n);
  
  for(i = 0; i < n; i++){
    j = i/DIM;
    k = i%DIM;
    uj = ((double) j - p[1])/p[3];
    uk = ((double) k - p[2])/p[4];
    x[i] = p[0]*exp(-0.5/(1.0 - p[5]*p[5])*(uj*uj + uk*uk - 2.0*p[5]*uj*uk));
  }
}

void jacgauss2d(double *p, double *jac, int m, int n, void *data)
{
  register int i, l;
  int j, k;
  double uj, uk;
  double tmp, corr;
  int DIM;
  
  DIM = (int) sqrt(n);

  for(i = l = 0; i < n; i++){
    j = i/DIM;
    k = i%DIM;
    uj = ((double) j - p[1])/p[3];
    uk = ((double) k - p[2])/p[4];
    corr = (1.0 - p[5]*p[5]);
    tmp = exp(-0.5/corr*(uj*uj + uk*uk - 2.0*p[5]*uj*uk));
    
    jac[l++] = tmp;
    jac[l++] = p[0]*tmp*(uj - p[5]*uk)/p[3]/corr;
    jac[l++] = p[0]*tmp*(uk - p[5]*uj)/p[4]/corr;
    jac[l++] = p[0]*tmp*(uj - p[5]*uk)*uj/p[3]/corr;
    jac[l++] = p[0]*tmp*(uk - p[5]*uj)*uk/p[4]/corr;
    jac[l++] = p[0]*tmp*(-(uj*uj + uk*uk)*p[5] + uj*uk*(1.0 + p[5]*p[5]))/corr/corr;
  }
}

sub_gcoef_LM(float *g, float *gcoef, int L)
{
  register int i;
  int ret;
  double p[6];
  double *x;
  double max_x = 0.0;
  int m, n;
  double opts[LM_OPTS_SZ], info[LM_INFO_SZ];
  int j, k;
  double uj, uk;

  int ii, jj;
  float mx = 0.0;
  float my = 0.0;
  float phi = 0.0;
  float lx = 0.0;
  float ly = 0.0;

  float k11 = 0.0;
  float k12 = 0.0;
  float k22 = 0.0;
  double theta = 0.0;
  float pi;
  float thres;

  float sum_g = 0.0;
  float sum_x = 0.0;
  float sum_y = 0.0;
  float sum_xx = 0.0;
  float sum_yy = 0.0;
  float sum_xy = 0.0;
  double sx = 0.0;
  double sy = 0.0;
  double rhoxy = 0.0;
  double mx_min, mx_max, my_min, my_max;
  
  opts[0] = LM_INIT_MU;
  opts[1] = 1E-15;
  opts[2] = 1E-15;
  opts[3] = LM_DIFF_DELTA; // relevant only if the no analytic jacobian version is used

  for (ii = 0; ii < L*L; ii++) sum_g += g[ii];

  if (sum_g > 0.0){
    for (ii = 0; ii < L; ii++){
      for (jj = 0; jj < L; jj++){
	sum_x += g[ii*L+jj]*((float) ii);
	sum_y += g[ii*L+jj]*((float) jj);
	sum_xx += g[ii*L+jj]*((float) (ii*ii));
	sum_yy += g[ii*L+jj]*((float) (jj*jj));
	sum_xy += g[ii*L+jj]*((float) (ii*jj));
      }
    }
    //calculate initial gaussian coefficients
    mx = sum_x/sum_g;
    my = sum_y/sum_g;
    k11 = sum_xx/sum_g - mx*mx;
    k22 = sum_yy/sum_g - my*my;
    k12 = sum_xy/sum_g - mx*my;
    sx = sqrt((double) k11);
    sy = sqrt((double) k22);
    rhoxy = ((double) k12)/sx/sy;

    pi = (float) acos(-1.0);
 
    m = 6; //number of coefficients
    n = L*L; //number of pixels per projection image
    x = (double *)malloc(n*sizeof(double));
    
    for(i = 0; i < n; i++){
      x[i] = (double) g[i];
      if (x[i] > max_x) max_x = x[i];
    }
    
    p[0] = max_x;
    p[1] = (double) mx;
    p[2] = (double) my;
    p[3] = sx;
    p[4] = sy;
    p[5] = rhoxy;
    
    mx_min = p[1] - 2.0*p[3];
    if (mx_min < 0.0) mx_min = 0.0;

    mx_max = p[1] + 2.0*p[3];
    if (mx_max > (double) (L - 1)) mx_max = (double) (L - 1);

    my_min = p[2] - 2.0*p[4];
    if (my_min < 0.0) my_min = 0.0;

    my_max = p[2] + 2.0*p[4];
    if (my_max > (double) (L - 1)) my_max = (double) (L - 1);

    //perform LS fit
    ret = dlevmar_der(gauss2d, jacgauss2d, p, x, m, n, 100, opts, info, NULL, NULL); // with analytic jacobian
    //ret = dlevmar_dif(gauss2d, p, x, m, n, 100, opts, info, NULL, NULL); // no jacobian

    //printf("%4.0f %g\n", max_x, info[5]);
    //printf("%.2f %.2f %.2f %.2f %.2f %.2f\n", sx, p[3], sy, p[4], rhoxy, p[5]);

    //assign the LS fit to the gaussian coefficients
    if ((info[5] < 100.0) && (p[1] > mx_min) && (p[1] < mx_max)
	&& (p[2] > my_min) && (p[2] < my_max)) {
      mx = (float) p[1];
      my = (float) p[2];
      //compute lx, ly, phi from sx, sy, and rhoxy
      k11 = (float) (p[3]*p[3]);
      k22 = (float) (p[4]*p[4]);
      k12 = (float) (p[3]*p[4]*p[5]);
    }
    
    thres = 1.0/((float) (L*L));
    //assume the standard deviation should be smaller than L and larger than 1/L
    
    if (k12 == 0.0){
      if (k11 >= k22){
	lx = k11;
	ly = k22;
	phi = 0.0;
      }
      else {
	lx = k22;
	ly = k11;
	phi = pi/2.0;
      }
    }
    else {
      if (k11 == k22){
	if (k12 < 0.0){
	  lx = k11 - k12;
	  ly = k11 + k12;
	  phi = pi/4.0;
	}
	else {
	  lx = k11 + k12;
	  ly = k11 - k12;
	  phi = -pi/4.0;
	}
      }
      else {
	theta = (0.5*atan2((double) (-2.0*k12),(double) (k11-k22)));
	lx = 0.5*(k11 + k22 - 2.0*k12/((float) sin(2.0*theta)));
	ly = 0.5*(k11 + k22 + 2.0*k12/((float) sin(2.0*theta)));
	phi = (float) theta; 
      }
    }
    //remove unreasonable lambda's
    if (lx <= thres) lx = 1.0;
    if (ly <= thres) ly = 1.0;

    //assign gcoef
    gcoef[0] = sum_g;
    gcoef[1] = mx;
    gcoef[2] = my;
    gcoef[3] = phi;
    gcoef[4] = lx;
    gcoef[5] = ly;
    
    free(x);
  }

  //if g_sum = 0.0
  else {
    for(ii = 0; ii < 6; ii++) gcoef[ii] = 0.0;
  }
}
